/* Copyright (C) 2013-2014 Michal Brzozowski (rusolis@poczta.fm)

   This file is part of KeeperRL.

   KeeperRL is free software; you can redistribute it and/or modify it under the terms of the
   GNU General Public License as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   KeeperRL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along with this program.
   If not, see http://www.gnu.org/licenses/ . */

#include "stdafx.h"

#include "creature.h"
#include "creature_factory.h"
#include "level.h"
#include "ranged_weapon.h"
#include "statistics.h"
#include "options.h"
#include "game.h"
#include "effect.h"
#include "item_factory.h"
#include "controller.h"
#include "player_message.h"
#include "attack.h"
#include "vision.h"
#include "equipment.h"
#include "shortest_path.h"
#include "spell_map.h"
#include "minion_task_map.h"
#include "tribe.h"
#include "creature_attributes.h"
#include "position.h"
#include "view.h"
#include "sound.h"
#include "lasting_effect.h"
#include "attack_type.h"
#include "attack_level.h"
#include "model.h"
#include "view_object.h"
#include "spell.h"
#include "body.h"
#include "field_of_view.h"
#include "furniture.h"
#include "creature_debt.h"

template <class Archive> 
void Creature::MoraleOverride::serialize(Archive& ar, const unsigned int version) {
}

SERIALIZABLE(Creature::MoraleOverride);

template <class Archive> 
void Creature::serialize(Archive& ar, const unsigned int version) { 
  ar & SUBCLASS(OwnedObject<Creature>) & SUBCLASS(Renderable) & SUBCLASS(UniqueEntity);
  ar(attributes, position, equipment, shortestPath, knownHiding, tribe, morale);
  ar(deathTime, hidden);
  ar(deathReason, swapPositionCooldown);
  ar(unknownAttackers, privateEnemies, holding);
  ar(controllerStack, creatureVisions, kills);
  ar(difficultyPoints, points, numAttacksThisTurn, moraleOverride);
  ar(vision, lastCombatTime, debt);
}

SERIALIZABLE(Creature);

SERIALIZATION_CONSTRUCTOR_IMPL(Creature);

Creature::Creature(const ViewObject& object, TribeId t, const CreatureAttributes& attr)
    : Renderable(object), attributes(attr), tribe(t) {
  modViewObject().setCreatureId(getUniqueId());
  updateVision();
}

Creature::Creature(TribeId t, const CreatureAttributes& attr)
    : Creature(attr.createViewObject(), t, attr) {
}

Creature::~Creature() {
}

vector<vector<WCreature>> Creature::stack(const vector<WCreature>& creatures) {
  map<string, vector<WCreature>> stacks;
  for (WCreature c : creatures)
    stacks[c->getName().stack()].push_back(c);
  return getValues(stacks);
}

const ViewObject& Creature::getViewObjectFor(const Tribe* observer) const {
  if (attributes->getIllusionViewObject() && observer->isEnemy(this))
    return *attributes->getIllusionViewObject();
  else
    return getViewObject();
}

const Body& Creature::getBody() const {
  return attributes->getBody();
}

Body& Creature::getBody() {
  return attributes->getBody();
}

double Creature::getSpellDelay(Spell* spell) const {
  CHECK(!isReady(spell));
  return attributes->getSpellMap().getReadyTime(spell) - getGlobalTime();
}

bool Creature::isReady(Spell* spell) const {
  return attributes->getSpellMap().getReadyTime(spell) < getGlobalTime();
}

static double getWillpowerMult(double sorcerySkill) {
  return 2 * pow(0.25, sorcerySkill); 
}

const CreatureAttributes& Creature::getAttributes() const {
  return *attributes;
}

CreatureAttributes& Creature::getAttributes() {
  return *attributes;
}

CreatureAction Creature::castSpell(Spell* spell) const {
  if (!attributes->getSpellMap().contains(spell))
    return CreatureAction("You don't know this spell.");
  CHECK(!spell->isDirected());
  if (!isReady(spell))
    return CreatureAction("You can't cast this spell yet.");
  return CreatureAction(this, [=] (WCreature c) {
    c->addSound(spell->getSound());
    spell->addMessage(c);
    Effect::applyToCreature(c, spell->getEffectType(), EffectStrength::NORMAL);
    getGame()->getStatistics().add(StatId::SPELL_CAST);
    c->attributes->getSpellMap().setReadyTime(spell, getGlobalTime() + spell->getDifficulty()
        * getWillpowerMult(attributes->getSkills().getValue(SkillId::SORCERY)));
    c->spendTime(1);
  });
}

CreatureAction Creature::castSpell(Spell* spell, Vec2 dir) const {
  CHECK(attributes->getSpellMap().contains(spell));
  CHECK(spell->isDirected());
  CHECK(dir.length8() == 1);
  if (!isReady(spell))
    return CreatureAction("You can't cast this spell yet.");
  return CreatureAction(this, [=] (WCreature c) {
    c->addSound(spell->getSound());
    monsterMessage(getName().the() + " casts a spell");
    playerMessage("You cast " + spell->getName());
    Effect::applyDirected(c, dir, spell->getDirEffectType(), EffectStrength::NORMAL);
    getGame()->getStatistics().add(StatId::SPELL_CAST);
    c->attributes->getSpellMap().setReadyTime(spell, getGlobalTime() + spell->getDifficulty()
        * getWillpowerMult(attributes->getSkills().getValue(SkillId::SORCERY)));
    c->spendTime(1);
  });
}

void Creature::addCreatureVision(WCreatureVision creatureVision) {
  creatureVisions.push_back(creatureVision);
}

void Creature::removeCreatureVision(WCreatureVision vision) {
  creatureVisions.removeElement(vision);
}

void Creature::pushController(PController ctrl) {
  if (ctrl->isPlayer())
    if (WGame g = getGame())
      g->setPlayer(this);
  controllerStack.push_back(std::move(ctrl));
}

void Creature::setController(PController ctrl) {
  controllerStack.clear();
  pushController(std::move(ctrl));
}

void Creature::popController() {
  if (auto controller = getController())
    if (controller->isPlayer())
      getGame()->clearPlayer();
  if (!controllerStack.empty()) {
    controllerStack.pop_back();
  }
}

bool Creature::isDead() const {
  return !!deathTime;
}

double Creature::getDeathTime() const {
  return *deathTime;
}

void Creature::clearLastAttacker() {
  lastAttacker = nullptr;
}

optional<string> Creature::getDeathReason() const {
  if (deathReason)
    return deathReason;
  if (lastAttacker)
    return "killed by " + lastAttacker->getName().a();
  return none;
}

const EntitySet<Creature>& Creature::getKills() const {
  return kills;
}

void Creature::spendTime(double t) {
  if (!isDead())
    if (WModel m = position.getModel())
      m->increaseLocalTime(this, 100.0 * t / (double) getAttr(AttrType::SPEED));
  hidden = false;
}

CreatureAction Creature::forceMove(Vec2 dir) const {
  return forceMove(getPosition().plus(dir));
}

CreatureAction Creature::forceMove(Position pos) const {
  const_cast<Creature*>(this)->forceMovement = true;
  CreatureAction action = move(pos);
  const_cast<Creature*>(this)->forceMovement = false;
  if (action)
    return action.prepend([] (WCreature c) { c->forceMovement = true; })
      .append([] (WCreature c) { c->forceMovement = false; });
  else
    return action;
}

CreatureAction Creature::move(Vec2 dir) const {
  return move(getPosition().plus(dir));
}

CreatureAction Creature::move(Position pos) const {
  Vec2 direction = getPosition().getDir(pos);
  if (getHoldingCreature())
    return CreatureAction("You can't break free!");
  if (direction.length8() != 1)
    return CreatureAction();
  if (!position.canMoveCreature(direction)) {
    if (pos.getCreature()) {
      if (!canSwapPositionInMovement(pos.getCreature()))
        return CreatureAction(/*"You can't swap position with " + pos.getCreature()->getName().the()*/);
    } else
      return CreatureAction();
  }
  return CreatureAction(this, [=](WCreature self) {
    INFO << getName().the() << " moving " << direction;
    if (isAffected(LastingEffect::ENTANGLED) || isAffected(LastingEffect::TIED_UP)) {
      playerMessage("You can't break free!");
      self->spendTime(1);
      return;
    }
    if (position.canMoveCreature(direction))
      self->position.moveCreature(direction);
    else {
      self->swapPosition(direction);
      return;
    }
    double oldTime = getLocalTime();
    if (getBody().isCollapsed(self)) {
      you(MsgType::CRAWL, getPosition().getName());
      self->spendTime(3);
    } else
      self->spendTime(1);
    self->addMovementInfo({direction, oldTime, getLocalTime(), MovementInfo::MOVE});
  });
}

void Creature::displace(double time, Vec2 dir) {
  position.moveCreature(dir);
  getController()->onDisplaced();
  addMovementInfo({dir, time, time + 1, MovementInfo::MOVE});
}

bool Creature::canTakeItems(const vector<WItem>& items) const {
  return getBody().isHumanoid() && canCarry(items);
}

void Creature::takeItems(vector<PItem> items, WCreature from) {
  vector<WItem> ref = getWeakPointers(items);
  equipment->addItems(std::move(items));
  getController()->onItemsGiven(ref, from);
}

void Creature::you(MsgType type, const vector<string>& param) const {
  getController()->you(type, param);
}

void Creature::you(MsgType type, const string& param) const {
  getController()->you(type, param);
}

void Creature::you(const string& param) const {
  getController()->you(param);
}

void Creature::playerMessage(const PlayerMessage& message) const {
  getController()->privateMessage(message);
}

WController Creature::getController() const {
  CHECK(!controllerStack.empty());
  return controllerStack.back().get();
}

bool Creature::hasCondition(CreatureCondition condition) const {
  for (auto effect : LastingEffects::getCausingCondition(condition))
    if (isAffected(effect))
      return true;
  return false;
}

bool Creature::canSwapPositionInMovement(WCreature other) const {
  return !other->hasCondition(CreatureCondition::RESTRICTED_MOVEMENT)
      && (swapPositionCooldown == 0 || isPlayer())
      && !other->getAttributes().isBoulder()
      && !other->isPlayer()
      && !other->isEnemy(this)
      && other->getPosition().canEnterEmpty(this)
      && getPosition().canEnterEmpty(other);
}

void Creature::swapPosition(Vec2 direction) {
  CHECK(direction.length8() == 1);
  WCreature other = NOTNULL(getPosition().plus(direction).getCreature());
  swapPositionCooldown = 4;
  playerMessage("Excuse me!");
  position.swapCreatures(other);
  double oldTime = getLocalTime();
  spendTime(1);
  addMovementInfo({direction, oldTime, getLocalTime(), MovementInfo::MOVE});
  other->addMovementInfo({-direction, oldTime, getLocalTime(), MovementInfo::MOVE});
}

void Creature::makeMove() {
  numAttacksThisTurn = 0;
  CHECK(!isDead());
  if (hasCondition(CreatureCondition::SLEEPING)) {
    getController()->sleeping();
    spendTime(1);
    return;
  }
  updateVisibleCreatures();
  updateViewObject();
  if (swapPositionCooldown)
    --swapPositionCooldown;
  {
    // Calls makeMove() while preventing Controller destruction by holding a shared_ptr on stack.
    // This is needed, otherwise Controller could be destroyed during makeMove() if creature committed suicide.
    shared_ptr<Controller> controllerTmp = controllerStack.back().giveMeSharedPointer();
    MEASURE(controllerTmp->makeMove(), "creature move time");
  }

  INFO << getName().bare() << " morale " << getMorale();
  if (!hidden)
    modViewObject().removeModifier(ViewObject::Modifier::HIDDEN);
  unknownAttackers.clear();
  getBody().affectPosition(position);
}

CreatureAction Creature::wait() const {
  return CreatureAction(this, [=](WCreature self) {
    INFO << getName().the() << " waiting";
    bool keepHiding = hidden;
    self->spendTime(1);
    self->hidden = keepHiding;
  });
}

const Equipment& Creature::getEquipment() const {
  return *equipment;
}

Equipment& Creature::getEquipment() {
  return *equipment;
}

vector<PItem> Creature::steal(const vector<WItem> items) {
  return equipment->removeItems(items, this);
}

WItem Creature::getAmmo() const {
  for (WItem item : equipment->getItems())
    if (item->getClass() == ItemClass::AMMO)
      return item;
  return nullptr;
}

WLevel Creature::getLevel() const {
  return getPosition().getLevel();
}

WGame Creature::getGame() const {
  return getPosition().getGame();
}

Position Creature::getPosition() const {
  return position;
}

void Creature::globalMessage(const PlayerMessage& playerCanSee) const {
  globalMessage(playerCanSee, "");
}

void Creature::globalMessage(const PlayerMessage& playerCanSee, const PlayerMessage& cant) const {
  position.globalMessage(this, playerCanSee, cant);
}

void Creature::monsterMessage(const PlayerMessage& playerCanSee, const PlayerMessage& cant) const {
  if (!isPlayer())
    position.globalMessage(this, playerCanSee, cant);
}

void Creature::monsterMessage(const PlayerMessage& playerCanSee) const {
  monsterMessage(playerCanSee, "");
}

void Creature::addSkill(Skill* skill) {
  if (!attributes->getSkills().hasDiscrete(skill->getId())) {
    attributes->getSkills().insert(skill->getId());
    playerMessage(skill->getHelpText());
  }
}

vector<WItem> Creature::getPickUpOptions() const {
  if (!getBody().isHumanoid())
    return vector<WItem>();
  else
    return getPosition().getItems();
}

string Creature::getPluralTheName(WItem item, int num) const {
  if (num == 1)
    return item->getTheName(false, this);
  else
    return toString(num) + " " + item->getTheName(true, this);
}

string Creature::getPluralAName(WItem item, int num) const {
  if (num == 1)
    return item->getAName(false, this);
  else
    return toString(num) + " " + item->getAName(true, this);
}

bool Creature::canCarry(const vector<WItem>& items) const {
  double weight = equipment->getTotalWeight();
  for (WItem it : items)
    weight += it->getWeight();
  return weight <= 2 * getModifier(ModifierType::INV_LIMIT);
}

CreatureAction Creature::pickUp(const vector<WItem>& items) const {
  if (!getBody().isHumanoid())
    return CreatureAction("You can't pick up anything!");
  if (!canCarry(items))
    return CreatureAction("You are carrying too much to pick this up.");
  return CreatureAction(this, [=](WCreature self) {
    INFO << getName().the() << " pickup ";
    for (auto stack : stackItems(items)) {
      monsterMessage(getName().the() + " picks up " + getPluralAName(stack[0], stack.size()));
      playerMessage("You pick up " + getPluralTheName(stack[0], stack.size()));
    }
    self->equipment->addItems(self->getPosition().removeItems(items));
    if (equipment->getTotalWeight() > getModifier(ModifierType::INV_LIMIT))
      playerMessage("You are overloaded.");
    getGame()->addEvent({EventId::PICKED_UP, EventInfo::ItemsHandled{self, items}});
    self->spendTime(1);
  });
}

vector<vector<WItem>> Creature::stackItems(vector<WItem> items) const {
  map<string, vector<WItem> > stacks = groupBy<WItem, string>(items, 
      [this] (WItem const& item) { return item->getNameAndModifiers(false, this); });
  return getValues(stacks);
}

CreatureAction Creature::drop(const vector<WItem>& items) const {
  if (!getBody().isHumanoid())
    return CreatureAction("You can't drop this item!");
  return CreatureAction(this, [=](WCreature self) {
    INFO << getName().the() << " drop";
    for (auto stack : stackItems(items)) {
      monsterMessage(getName().the() + " drops " + getPluralAName(stack[0], stack.size()));
      playerMessage("You drop " + getPluralTheName(stack[0], stack.size()));
    }
    getGame()->addEvent({EventId::DROPPED, EventInfo::ItemsHandled{self, items}});
    self->getPosition().dropItems(self->equipment->removeItems(items, self));
    self->spendTime(1);
  });
}

void Creature::drop(vector<PItem> items) {
  getPosition().dropItems(std::move(items));
}

bool Creature::canEquipIfEmptySlot(const WItem item, string* reason) const {
  if (!getBody().isHumanoid()) {
    if (reason)
      *reason = "Only humanoids can equip items!";
    return false;
  }
  if (!attributes->canEquip()) {
    if (reason)
      *reason = "You can't equip items!";
    return false;
  }
  if (getBody().numGood(BodyPart::ARM) == 0) {
    if (reason)
      *reason = "You have no healthy arms!";
    return false;
  }
  if (getBody().numGood(BodyPart::ARM) == 1 && item->isWieldedTwoHanded()) {
    if (reason)
      *reason = "You need two hands to wield " + item->getAName() + "!";
    return false;
  }
  return item->canEquip();
}

bool Creature::canEquip(const WItem item) const {
  return canEquipIfEmptySlot(item, nullptr) && equipment->canEquip(item);
}

bool Creature::isEquipmentAppropriate(const WItem item) const {
  return item->getClass() != ItemClass::WEAPON || item->getMinStrength() <= getAttr(AttrType::STRENGTH);
}

CreatureAction Creature::equip(WItem item) const {
  string reason;
  if (!canEquipIfEmptySlot(item, &reason))
    return CreatureAction(reason);
  if (equipment->getSlotItems(item->getEquipmentSlot()).contains(item))
    return CreatureAction();
  return CreatureAction(this, [=](WCreature self) {
    INFO << getName().the() << " equip " << item->getName();
    EquipmentSlot slot = item->getEquipmentSlot();
    if (self->equipment->getSlotItems(slot).size() >= self->equipment->getMaxItems(slot)) {
      WItem previousItem = self->equipment->getSlotItems(slot)[0];
      self->equipment->unequip(previousItem, self);
    }
    self->equipment->equip(item, slot, self);
    playerMessage("You equip " + item->getTheName(false, self));
    monsterMessage(getName().the() + " equips " + item->getAName());
    if (WGame game = getGame())
      game->addEvent({EventId::EQUIPED, EventInfo::ItemsHandled{self, {item}}});
    self->spendTime(1);
  });
}

CreatureAction Creature::unequip(WItem item) const {
  if (!equipment->isEquipped(item))
    return CreatureAction("This item is not equipped.");
  if (!getBody().isHumanoid())
    return CreatureAction("You can't remove this item!");
  if (getBody().numGood(BodyPart::ARM) == 0)
    return CreatureAction("You have no healthy arms!");
  return CreatureAction(this, [=](WCreature self) {
    INFO << getName().the() << " unequip";
    CHECK(equipment->isEquipped(item)) << "Item not equipped.";
    EquipmentSlot slot = item->getEquipmentSlot();
    self->equipment->unequip(item, self);
    playerMessage("You " + string(slot == EquipmentSlot::WEAPON ? " sheathe " : " remove ") +
        item->getTheName(false, this));
    monsterMessage(getName().the() + (slot == EquipmentSlot::WEAPON ? " sheathes " : " removes ") +
        item->getAName());
    self->spendTime(1);
  });
}

CreatureAction Creature::heal(Vec2 direction) const {
  WConstCreature other = getPosition().plus(direction).getCreature();
  if (!attributes->getSkills().hasDiscrete(SkillId::HEALING) || !other || !other->getBody().canHeal() ||
      other == this)
    return CreatureAction();
  return CreatureAction(this, [=](WCreature self) {
    WCreature other = getPosition().plus(direction).getCreature();
    other->playerMessage("\"Let me help you my friend.\"");
    other->you(MsgType::ARE, "healed by " + getName().the());
    other->heal();
    self->spendTime(1);
  });
}

CreatureAction Creature::bumpInto(Vec2 direction) const {
  if (WConstCreature other = getPosition().plus(direction).getCreature())
    return CreatureAction(this, [=](WCreature self) {
      other->getController()->onBump(self);
      self->spendTime(1);
    });
  else
    return CreatureAction();
}

CreatureAction Creature::applySquare(Position pos) const {
  CHECK(pos.dist8(getPosition()) <= 1);
  if (auto furniture = pos.getFurniture(FurnitureLayer::MIDDLE))
    if (furniture->canUse(this))
      return CreatureAction(this, [=](WCreature self) {
        INFO << getName().the() << " applying " << getPosition().getName();
        auto originalPos = getPosition();
        double usageTime = furniture->getUsageTime();
        furniture->use(pos, self);
        double oldTime = getLocalTime();
        self->spendTime(usageTime);
        if (pos != getPosition() && getPosition() == originalPos)
          self->addMovementInfo({getPosition().getDir(pos), oldTime, min(oldTime + 1, getLocalTime()),
              MovementInfo::ATTACK});
      });
  return CreatureAction();
}

CreatureAction Creature::hide() const {
  if (!attributes->getSkills().hasDiscrete(SkillId::AMBUSH))
    return CreatureAction("You don't have this skill.");
  if (auto furniture = getPosition().getFurniture(FurnitureLayer::MIDDLE))
    if (furniture->canHide())
      return CreatureAction(this, [=](WCreature self) {
        playerMessage("You hide behind the " + furniture->getName());
        self->knownHiding.clear();
        self->modViewObject().setModifier(ViewObject::Modifier::HIDDEN);
        for (WCreature other : getLevel()->getAllCreatures())
          if (other->canSee(this) && other->isEnemy(this)) {
            self->knownHiding.insert(other);
            if (!isBlind())
              you(MsgType::CAN_SEE_HIDING, other->getName().the());
          }
        self->spendTime(1);
        self->hidden = true;
      });
  return CreatureAction("You can't hide here.");
}

CreatureAction Creature::chatTo(WCreature other) const {
  CHECK(other);
  if (other->getPosition().dist8(getPosition()) == 1)
    return CreatureAction(this, [=](WCreature self) {
        playerMessage("You chat with " + other->getName().the());
        other->getAttributes().chatReaction(other, self);
        self->spendTime(1);
    });
  else
    return CreatureAction("Move closer to chat to " + other->getName().the());
}

CreatureAction Creature::stealFrom(Vec2 direction, const vector<WItem>& items) const {
  if (getPosition().plus(direction).getCreature())
    return CreatureAction(this, [=](WCreature self) {
        WCreature other = NOTNULL(getPosition().plus(direction).getCreature());
        self->equipment->addItems(other->steal(items));
      });
  return CreatureAction();
}

bool Creature::isHidden() const {
  return hidden;
}

bool Creature::knowsHiding(WConstCreature c) const {
  return knownHiding.contains(c);
}

void Creature::addEffect(LastingEffect effect, double time, bool msg) {
  if (LastingEffects::affects(this, effect) && attributes->considerAffecting(effect, getGlobalTime(), time))
    LastingEffects::onAffected(this, effect, msg);
}

void Creature::removeEffect(LastingEffect effect, bool msg) {
  if (!isAffected(effect))
    return;
  attributes->clearLastingEffect(effect);
  if (!isAffected(effect))
    LastingEffects::onRemoved(this, effect, msg);
}

void Creature::addPermanentEffect(LastingEffect effect, bool msg) {
  if (!isAffected(effect))
    LastingEffects::onAffected(this, effect, msg);
  attributes->addPermanentEffect(effect);
}

void Creature::removePermanentEffect(LastingEffect effect, bool msg) {
  attributes->removePermanentEffect(effect);
  if (!isAffected(effect))
    LastingEffects::onRemoved(this, effect, msg);
}

bool Creature::isAffected(LastingEffect effect) const {
  return attributes->isAffected(effect, getGlobalTime());
}

optional<double> Creature::getTimeRemaining(LastingEffect effect) const {
  double t = attributes->getTimeOut(effect);
  double global = getGlobalTime();
  if (t >= global)
    return t - global;
  else
    return none;
}

bool Creature::isBlind() const {
  return isAffected(LastingEffect::BLIND) ||
    (getBody().numLost(BodyPart::HEAD) > 0 && getBody().numBodyParts(BodyPart::HEAD) == 0);
}

bool Creature::isDarknessSource() const {
  return isAffected(LastingEffect::DARKNESS_SOURCE);
}

// penalty to strength and dexterity per extra attacker in a single turn
int simulAttackPen(int attackers) {
  return max(0, (attackers - 1) * 2);
}

int Creature::getAttr(AttrType type) const {
  int def = getBody().modifyAttr(type, attributes->getRawAttr(type));
  for (WItem item : equipment->getAllEquipped())
    def += item->getAttr(type);
  switch (type) {
    case AttrType::DEXTERITY:
    case AttrType::STRENGTH:
        def -= simulAttackPen(numAttacksThisTurn);
        break;
    case AttrType::SPEED: {
        double totWeight = equipment->getTotalWeight();
        if (!attributes->canCarryAnything() && totWeight > getAttr(AttrType::STRENGTH))
          def -= 20.0 * totWeight / def;
        CHECK(def > 0);
        break;}
  }
  LastingEffects::modifyAttr(this, type, def);
  return max(0, def);
}

int Creature::accuracyBonus() const {
  if (WItem weapon = getWeapon())
    return -max(0, weapon->getMinStrength() - getAttr(AttrType::STRENGTH));
  else
    return 0;
}

int Creature::getModifier(ModifierType type) const {
  int def = 0;
  for (WItem item : equipment->getAllEquipped())
      def += item->getModifier(type);
  for (SkillId skill : ENUM_ALL(SkillId))
    def += Skill::get(skill)->getModifier(this, type);
  switch (type) {
    case ModifierType::FIRED_DAMAGE: 
    case ModifierType::THROWN_DAMAGE: 
        def += getAttr(AttrType::DEXTERITY);
        break;
    case ModifierType::DAMAGE: 
        def += getAttr(AttrType::STRENGTH);
        if (!getWeapon())
          def += attributes->getBarehandedDamage();
        break;
    case ModifierType::DEFENSE: 
        def += getAttr(AttrType::STRENGTH);
        break;
    case ModifierType::FIRED_ACCURACY: 
    case ModifierType::THROWN_ACCURACY: 
        def += getAttr(AttrType::DEXTERITY);
        break;
    case ModifierType::ACCURACY: 
        def += accuracyBonus();
        def += getAttr(AttrType::DEXTERITY);
        break;
    case ModifierType::INV_LIMIT:
        if (attributes->canCarryAnything())
          return 1000000;
        return getAttr(AttrType::STRENGTH) * 2;
  }
  LastingEffects::modifyMod(this, type, def);
  return max(0, def);
}

int Creature::getPoints() const {
  return points;
}

void Creature::onKilled(WCreature victim) {
  int difficulty = victim->getDifficultyPoints();
  CHECK(difficulty >=0 && difficulty < 100000) << difficulty << " " << victim->getName().bare();
  points += difficulty;
  increaseExpLevel(ExperienceType::COMBAT, getAttributes().getExpFromKill(victim));
  kills.insert(victim);
}

Tribe* Creature::getTribe() {
  return getGame()->getTribe(tribe);
}

const Tribe* Creature::getTribe() const {
  return getGame()->getTribe(tribe);
}

TribeId Creature::getTribeId() const {
  return tribe;
}

void Creature::setTribe(TribeId t) {
  tribe = t;
}

bool Creature::isFriend(WConstCreature c) const {
  return !isEnemy(c);
}

bool Creature::isEnemy(WConstCreature c) const {
  if (c == this)
    return false;
  if (isAffected(LastingEffect::INSANITY))
    return c != this;
  return getTribe()->isEnemy(c) || c->getTribe()->isEnemy(this) ||
    privateEnemies.contains(c) || c->privateEnemies.contains(this);
}

vector<WItem> Creature::getGold(int num) const {
  vector<WItem> ret;
  for (WItem item : equipment->getItems([](WItem it) { return it->getClass() == ItemClass::GOLD; })) {
    ret.push_back(item);
    if (ret.size() == num)
      return ret;
  }
  return ret;
}

void Creature::setPosition(Position pos) {
  if (!pos.isSameLevel(position)) {
    modViewObject().clearMovementInfo();
  }
  if (shortestPath && shortestPath->getLevel() != pos.getLevel())
    shortestPath.reset();
  position = pos;
}

double Creature::getLocalTime() const {
  if (WModel m = position.getModel())
    return m->getLocalTime(this);
  else
    return 0;
}

double Creature::getGlobalTime() const {
  if (WGame g = getGame())
    return g->getGlobalTime();
  else
    return 1;
}

void Creature::tick() {
  updateVision();
  if (Random.roll(5))
    getDifficultyPoints();
  for (WItem item : equipment->getItems()) {
    item->tick(position);
    if (item->isDiscarded())
      equipment->removeItem(item, this);
  }
  double globalTime = getGlobalTime();
  for (LastingEffect effect : ENUM_ALL(LastingEffect))
    if (attributes->considerTimeout(effect, globalTime))
      LastingEffects::onTimedOut(this, effect, true);
  if (isAffected(LastingEffect::POISON))
    if (getBody().affectByPoison(this, 0.015)) {
      dieWithAttacker(lastAttacker);
      return;
    }
  updateViewObject();
  if (getBody().tick(this)) {
    dieWithAttacker(lastAttacker);
    return;
  }
}

static string getAttackParam(AttackType type) {
  switch (type) {
    case AttackType::CUT: return "cut";
    case AttackType::STAB: return "stab";
    case AttackType::CRUSH: return "crush";
    case AttackType::PUNCH: return "punch";
    case AttackType::EAT:
    case AttackType::BITE: return "bite";
    case AttackType::HIT: return "hit";
    case AttackType::SHOOT: return "shot";
    case AttackType::SPELL: return "spell";
    case AttackType::POSSESS: return "touch";
  }
}

string Creature::getAttrName(AttrType attr) {
  switch (attr) {
    case AttrType::STRENGTH: return "strength";
    case AttrType::DEXTERITY: return "dexterity";
    case AttrType::SPEED: return "speed";
  }
}

string Creature::getModifierName(ModifierType attr) {
  switch (attr) {
    case ModifierType::DAMAGE: return "damage";
    case ModifierType::ACCURACY: return "accuracy";
    case ModifierType::THROWN_DAMAGE: return "thrown damage";
    case ModifierType::THROWN_ACCURACY: return "thrown accuracy";
    case ModifierType::FIRED_DAMAGE: return "projectile damage";
    case ModifierType::FIRED_ACCURACY: return "projectile accuracy";
    case ModifierType::DEFENSE: return "defense";
    case ModifierType::INV_LIMIT: return "carry capacity";
  }
}

static MsgType getAttackMsg(AttackType type, bool weapon, AttackLevel level) {
  if (weapon)
    return type == AttackType::STAB ? MsgType::THRUST_WEAPON : MsgType::SWING_WEAPON;
  switch (type) {
    case AttackType::EAT:
    case AttackType::BITE: return MsgType::BITE;
    case AttackType::PUNCH: return level == AttackLevel::LOW ? MsgType::KICK : MsgType::PUNCH;
    case AttackType::HIT: return MsgType::HIT;
    case AttackType::POSSESS: return MsgType::TOUCH;
    default: FATAL << "Unhandled barehanded attack: " << int(type);
  }
  return MsgType(0);
}

void Creature::dropWeapon() {
  if (auto weapon = getWeapon()) {
    you(MsgType::DROP_WEAPON, weapon->getName());
    getPosition().dropItem(equipment->removeItem(weapon, this));
  }
}

CreatureAction Creature::attack(WCreature other, optional<AttackParams> attackParams, bool spend) const {
  CHECK(!other->isDead());
  if (!position.isSameLevel(other->getPosition()))
    return CreatureAction();
  Vec2 dir = getPosition().getDir(other->getPosition());
  if (dir.length8() != 1)
    return CreatureAction();
  return CreatureAction(this, [=] (WCreature c) {
  INFO << getName().the() << " attacking " << other->getName().the();
  int accuracy = getModifier(ModifierType::ACCURACY);
  int damage = getModifier(ModifierType::DAMAGE);
  int accuracyVariance = 1 + accuracy / 3;
  int damageVariance = 1 + damage / 3;
  auto rAccuracy = [=] () { return Random.get(-accuracyVariance, accuracyVariance); };
  auto rDamage = [=] () { return Random.get(-damageVariance, damageVariance); };
  double timeSpent = 1;
  accuracy += rAccuracy() + rAccuracy();
  damage += rDamage() + rDamage();
  vector<string> attackAdjective;
  if (attackParams && attackParams->mod)
    switch (*attackParams->mod) {
      case AttackParams::WILD: 
        damage *= 1.2;
        accuracy *= 0.8;
        timeSpent *= 1.5;
        attackAdjective.push_back("wildly");
        break;
      case AttackParams::SWIFT: 
        damage *= 0.8;
        accuracy *= 1.2;
        timeSpent *= 0.7;
        attackAdjective.push_back("swiftly");
        break;
    }
  bool backstab = false;
  string enemyName = getLevel()->playerCanSee(other) ? other->getName().the() : "something";
  if (other->isPlayer())
    enemyName = "";
  if (!other->canSee(this) && canSee(other)) {
 //   if (getWeapon() && getWeapon()->getAttackType() == AttackType::STAB) {
      damage += 10;
      backstab = true;
 //   }
    you(MsgType::ATTACK_SURPRISE, enemyName);
  }
  AttackLevel attackLevel = Random.choose(getBody().getAttackLevels());
  if (attackParams && attackParams->level)
    attackLevel = *attackParams->level;
  Attack attack(c, attackLevel, attributes->getAttackType(getWeapon()), accuracy, damage, backstab,
      getWeapon() ? getWeapon()->getAttackEffect() : attributes->getAttackEffect());
  if (!other->dodgeAttack(attack)) {
    if (getWeapon()) {
      you(getAttackMsg(attack.getType(), true, attack.getLevel()),
          concat({getWeapon()->getName()}, attackAdjective));
      if (!canSee(other))
        playerMessage("You hit something.");
    } else
      you(getAttackMsg(attack.getType(), false, attack.getLevel()), concat({enemyName}, attackAdjective));
    other->takeDamage(attack);
  } else {
    you(MsgType::MISS_ATTACK, enemyName);
    addSound(SoundId::MISSED_ATTACK);
  }
  double oldTime = getLocalTime();
  if (spend)
    c->spendTime(timeSpent);
  c->addMovementInfo({dir, oldTime, getLocalTime(), MovementInfo::ATTACK});
  });
}

bool Creature::dodgeAttack(const Attack& attack) {
  ++numAttacksThisTurn;
  WCreature attacker = attack.getAttacker();
  if (attacker) {
    if (!canSee(attacker))
      unknownAttackers.insert(attacker);
  }
  return (!attacker || canSee(attacker)) && attack.getAccuracy() <= getModifier(ModifierType::ACCURACY);
}

bool Creature::takeDamage(const Attack& attack) {
  AttackType attackType = attack.getType();
  int defense = getModifier(ModifierType::DEFENSE);
  if (WCreature attacker = attack.getAttacker()) {
    if (attacker->tribe != tribe || Random.roll(3))
      privateEnemies.insert(attacker);
    if (!attacker->getAttributes().getSkills().hasDiscrete(SkillId::STEALTH))
      for (Position p : visibleCreatures)
        if (p.dist8(position) < 10 && p.getCreature() && !p.getCreature()->isDead())
          p.getCreature()->removeEffect(LastingEffect::SLEEP);
    if (attackType == AttackType::POSSESS) {
      you(MsgType::ARE, "possessed by " + attacker->getName().the());
      attacker->dieNoReason(Creature::DropType::NOTHING);
      addEffect(LastingEffect::INSANITY, 10);
      return false;
    }
    INFO << getName().the() << " attacked by " << attacker->getName().the()
      << " damage " << attack.getStrength() << " defense " << defense;
    lastAttacker = attack.getAttacker();
  }
  if (auto sound = attributes->getAttackSound(attack.getType(), attack.getStrength() > defense))
    addSound(*sound);
  if (attack.getStrength() > defense) {
    double dam = (defense == 0) ? 1 : double(attack.getStrength() - defense) / defense;
    if (attributes->getBody().takeDamage(attack, this, dam))
      return true;
  } else {
    you(MsgType::GET_HIT_NODAMAGE, getAttackParam(attackType));
    if (attack.getEffect())
      Effect::applyToCreature(this, *attack.getEffect(), EffectStrength::NORMAL);
  }
  for (LastingEffect effect : ENUM_ALL(LastingEffect))
    if (isAffected(effect))
      LastingEffects::onCreatureDamage(this, effect);
  return false;
}

static vector<string> extractNames(const vector<AdjectiveInfo>& adjectives) {
  return adjectives.transform([] (const AdjectiveInfo& e) -> string { return e.name; });
}

void Creature::updateViewObject() {
  modViewObject().setAttribute(ViewObject::Attribute::DEFENSE, getModifier(ModifierType::DEFENSE));
  modViewObject().setAttribute(ViewObject::Attribute::ATTACK, getModifier(ModifierType::DAMAGE));
  modViewObject().setAttribute(ViewObject::Attribute::LEVEL, attributes->getVisibleExpLevel());
  modViewObject().setAttribute(ViewObject::Attribute::MORALE, getMorale());
  modViewObject().setModifier(ViewObject::Modifier::DRAW_MORALE);
  modViewObject().setAdjectives(extractNames(concat(
          getWeaponAdjective(), getBadAdjectives(), getGoodAdjectives())));
  getBody().updateViewObject(modViewObject());
  modViewObject().setDescription(getName().bare());
  getPosition().setNeedsRenderUpdate(true);
}

double Creature::getMorale() const {
  if (moraleOverride)
    if (auto ret = moraleOverride->getMorale(this))
      return *ret;
  return morale;
}

void Creature::addMorale(double val) {
  morale = min(1.0, max(-1.0, morale + val));
}

void Creature::setMoraleOverride(PMoraleOverride mod) {
  moraleOverride = std::move(mod);
}

string attrStr(bool strong, bool agile, bool fast) {
  vector<string> good;
  vector<string> bad;
  if (strong)
    good.push_back("strong");
  else
    bad.push_back("weak");
  if (agile)
    good.push_back("agile");
  else
    bad.push_back("clumsy");
  if (fast)
    good.push_back("fast");
  else
    bad.push_back("slow");
  string p1 = combine(good);
  string p2 = combine(bad);
  if (p1.size() > 0 && p2.size() > 0)
    p1.append(", but ");
  p1.append(p2);
  return p1;
}

void Creature::heal(double amount, bool replaceLimbs) {
  if (getBody().heal(this, amount, replaceLimbs))
    clearLastAttacker();
  updateViewObject();
}

void Creature::fireDamage(double amount) {
  if (!isAffected(LastingEffect::FIRE_RESISTANT))
    getBody().fireDamage(this, amount);
}

void Creature::affectBySilver() {
  if (getBody().affectBySilver(this))
    dieWithAttacker(lastAttacker);
}

void Creature::affectByAcid() {
  if (getBody().affectByAcid(this))
    dieWithReason("dissolved by acid");
}

void Creature::poisonWithGas(double amount) {
  if (getBody().affectByPoisonGas(this, amount))
    dieWithReason("poisoned with gas");
}

void Creature::setHeld(WCreature c) {
  holding = c->getUniqueId();
}

WCreature Creature::getHoldingCreature() const {
  if (holding)
    for (auto pos : getPosition().neighbors8())
      if (auto c = pos.getCreature())
        if (c->getUniqueId() == *holding)
          return c;
  return nullptr;
}

void Creature::take(vector<PItem> items) {
  for (PItem& elem : items)
    take(std::move(elem));
}

void Creature::take(PItem item) {
  WItem ref = item.get();
  equipment->addItem(std::move(item));
  if (auto action = equip(ref))
    action.perform(this);
}

void Creature::dieWithReason(const string& reason, DropType drops) {
  deathReason = reason;
  dieNoReason(drops);
}

void Creature::dieWithAttacker(WCreature attacker, DropType drops) {
  CHECK(!isDead()) << getName().bare() << " is already dead. " << getDeathReason().value_or("");
  deathTime = getGlobalTime();
  lastAttacker = attacker;
  INFO << getName().the() << " dies. Killed by " << (attacker ? attacker->getName().bare() : "");
  getController()->onKilled(attacker);
  if (drops == DropType::EVERYTHING || drops == DropType::ONLY_INVENTORY)
    for (PItem& item : equipment->removeAllItems(this))
      getPosition().dropItem(std::move(item));
  if (drops == DropType::EVERYTHING) {
    getPosition().dropItems(getBody().getCorpseItem(getName().bare(), getUniqueId()));
    if (auto sound = getBody().getDeathSound())
      addSound(*sound);
  }
  if (attributes->isInnocent())
    getGame()->getStatistics().add(StatId::INNOCENT_KILLED);
  getGame()->getStatistics().add(StatId::DEATH);
  if (attacker)
    attacker->onKilled(this);
  getGame()->addEvent({EventId::KILLED, EventInfo::Attacked{this, attacker}});
  getTribe()->onMemberKilled(this, attacker);
  getLevel()->killCreature(this);
  setController(makeOwner<DoNothingController>(this));
}

void Creature::dieNoReason(DropType drops) {
  dieWithAttacker(nullptr, drops);
}

CreatureAction Creature::flyAway() const {
  if (!isAffected(LastingEffect::FLYING) || getPosition().isCovered())
    return CreatureAction();
  return CreatureAction(this, [=](WCreature self) {
    INFO << getName().the() << " fly away";
    monsterMessage(getName().the() + " flies away.");
    self->dieNoReason(Creature::DropType::NOTHING);
  });
}

CreatureAction Creature::disappear() const {
  return CreatureAction(this, [=](WCreature self) {
    INFO << getName().the() << " disappears";
    monsterMessage(getName().the() + " disappears.");
    self->dieNoReason(Creature::DropType::NOTHING);
  });
}

CreatureAction Creature::torture(WCreature other) const {
  if (!other->hasCondition(CreatureCondition::RESTRICTED_MOVEMENT) || other->getPosition().dist8(getPosition()) != 1)
    return CreatureAction();
  return CreatureAction(this, [=](WCreature self) {
    monsterMessage(getName().the() + " tortures " + other->getName().the());
    playerMessage("You torture " + other->getName().the());
    if (Random.roll(4))
      other->monsterMessage(other->getName().the() + " screams!", "You hear a horrible scream");
    other->addEffect(LastingEffect::STUNNED, 3, false);
    other->getBody().affectByTorture(other);
    getGame()->addEvent({EventId::TORTURED, EventInfo::Attacked{other, self}});
    self->spendTime(1);
  });
}

void Creature::surrender(WCreature to) {
  getGame()->addEvent({EventId::SURRENDERED, EventInfo::Attacked{this, to}});
}

void Creature::retire() {
  if (auto id = attributes->getRetiredViewId())
    modViewObject().setId(*id);
}

void Creature::increaseExpLevel(ExperienceType type, double increase) {
  int curLevel = (int)getAttributes().getVisibleExpLevel();
  getAttributes().increaseExpLevel(type, increase);
  int newLevel = (int)getAttributes().getVisibleExpLevel();
  if (curLevel != newLevel) {
    you(MsgType::ARE, "more experienced");
    addPersonalEvent(getName().a() + " reaches experience level " + toString(newLevel));
  }
}

CreatureAction Creature::give(WCreature whom, vector<WItem> items) const {
  if (!getBody().isHumanoid() || !whom->canTakeItems(items))
    return CreatureAction(whom->getName().the() + (items.size() == 1 ? " can't take this item."
        : " can't take these items."));
  return CreatureAction(this, [=](WCreature self) {
    for (auto stack : stackItems(items)) {
      if (!whom->isPlayer())
        monsterMessage(getName().the() + " gives " + getPluralAName(stack[0], (int) stack.size()) + " to " +
            whom->getName().the());
      whom->playerMessage(getName().the() + " gives you " + getPluralAName(stack[0], (int) stack.size()));
      playerMessage("You give " + getPluralTheName(stack[0], (int) stack.size()) + " to " +
        whom->getName().the());
    }
    whom->takeItems(self->equipment->removeItems(items, self), self);
  });
}

CreatureAction Creature::payFor(const vector<WItem>& items) const {
  int totalPrice = std::accumulate(items.begin(), items.end(), 0,
      [](int sum, const WItem it) { return sum + it->getPrice(); });
  return give(items[0]->getShopkeeper(this), getGold(totalPrice))
      .append([=](WCreature) { for (auto it : items) it->setShopkeeper(nullptr); });
}

CreatureAction Creature::fire(Vec2 direction) const {
  CHECK(direction.length8() == 1);
  if (getEquipment().getItems(ItemIndex::RANGED_WEAPON).empty())
    return CreatureAction("You need a ranged weapon.");
  if (getEquipment().getSlotItems(EquipmentSlot::RANGED_WEAPON).empty())
    return CreatureAction("You need to equip your ranged weapon.");
  if (getBody().numGood(BodyPart::ARM) < 2)
    return CreatureAction("You need two hands to shoot a bow.");
  if (!getAmmo())
    return CreatureAction("Out of ammunition");
  return CreatureAction(this, [=](WCreature self) {
    PItem ammo = self->equipment->removeItem(NOTNULL(getAmmo()), self);
    auto weapon = self->getEquipment().getSlotItems(EquipmentSlot::RANGED_WEAPON).getOnlyElement()
        .dynamicCast<RangedWeapon>();
    CHECK(!!weapon);
    weapon->fire(self, std::move(ammo), direction);
    self->spendTime(1);
  });
}

void Creature::addMovementInfo(const MovementInfo& info) {
  modViewObject().addMovementInfo(info);
  getPosition().setNeedsRenderUpdate(true);
}

CreatureAction Creature::whip(const Position& pos) const {
  WCreature whipped = pos.getCreature();
  if (pos.dist8(position) > 1 || !whipped)
    return CreatureAction();
  return CreatureAction(this, [=](WCreature self) {
    monsterMessage(PlayerMessage(getName().the() + " whips " + whipped->getName().the()));
    double oldTime = getLocalTime();
    self->spendTime(1);
    if (Random.roll(3)) {
      addSound(SoundId::WHIP);
      self->addMovementInfo({position.getDir(pos), oldTime, getLocalTime(), MovementInfo::ATTACK});
    }
    if (Random.roll(5))
      whipped->monsterMessage(whipped->getName().the() + " screams!", "You hear a horrible scream!");
    if (Random.roll(10)) {
      whipped->addMorale(0.05);
      whipped->you(MsgType::FEEL, "happier");
    }
  });
}

void Creature::addSound(const Sound& sound1) const {
  Sound sound(sound1);
  sound.setPosition(getPosition());
  getGame()->getView()->addSound(sound);
}

CreatureAction Creature::construct(Vec2 direction, FurnitureType type) const {
  if (getPosition().plus(direction).canConstruct(type) && canConstruct(type))
    return CreatureAction(this, [=](WCreature self) {
        addSound(Sound(SoundId::DIGGING).setPitch(0.5));
        getPosition().plus(direction).construct(type, self);
        self->spendTime(1);
      });
  return CreatureAction();
}

bool Creature::canConstruct(FurnitureType type) const {
  return attributes->getSkills().hasDiscrete(SkillId::CONSTRUCTION);
}

CreatureAction Creature::eat(WItem item) const {
  return CreatureAction(this, [=](WCreature self) {
    monsterMessage(getName().the() + " eats " + item->getAName());
    playerMessage("You eat " + item->getAName());
    self->getPosition().removeItem(item);
    self->spendTime(3);
  });
}

void Creature::destroyImpl(Vec2 direction, const DestroyAction& action) {
  auto pos = getPosition().plus(direction);
  if (auto furniture = pos.modFurniture(FurnitureLayer::MIDDLE)) {
    string name = furniture->getName();
    playerMessage("You "_s + action.getVerbSecondPerson() + " the " + name);
    monsterMessage(getName().the() + " " + action.getVerbThirdPerson() + " the " + name, action.getSoundText());
    furniture->tryToDestroyBy(pos, this, action);
  }
}

CreatureAction Creature::destroy(Vec2 direction, const DestroyAction& action) const {
  auto pos = getPosition().plus(direction);
  if (auto furniture = pos.getFurniture(FurnitureLayer::MIDDLE))
    if (direction.length8() <= 1 && furniture->canDestroy(getMovementType(), action))
      return CreatureAction(this, [=](WCreature self) {
        self->destroyImpl(direction, action);
        double oldTime = getLocalTime();
        self->spendTime(1);
        if (direction.length8() == 1)
          self->addMovementInfo({getPosition().getDir(pos), oldTime, min(oldTime + 1, getLocalTime()),
              MovementInfo::ATTACK});
      });
  return CreatureAction();
}

bool Creature::canCopulateWith(WConstCreature c) const {
  return c->getBody().canCopulateWith() && c->attributes->getGender() != attributes->getGender() &&
    c->isAffected(LastingEffect::SLEEP);
}

bool Creature::canConsume(WConstCreature c) const {
  return c->getBody().canConsume() && attributes->getSkills().hasDiscrete(SkillId::CONSUMPTION) && isFriend(c);
}

CreatureAction Creature::copulate(Vec2 direction) const {
  WConstCreature other = getPosition().plus(direction).getCreature();
  if (!other || !canCopulateWith(other))
    return CreatureAction();
  return CreatureAction(this, [=](WCreature self) {
      INFO << getName().bare() << " copulate with " << other->getName().bare();
      you(MsgType::COPULATE, "with " + other->getName().the());
      self->spendTime(2);
    });
}

void Creature::addPersonalEvent(const string& s) {
  if (WModel m = position.getModel())
    m->addEvent({EventId::CREATURE_EVENT, EventInfo::CreatureEvent{this, s}});
}

CreatureAction Creature::consume(WCreature other) const {
  if (!other || !canConsume(other) || other->getPosition().dist8(getPosition()) > 1)
    return CreatureAction();
  return CreatureAction(this, [=] (WCreature self) {
    self->attributes->consume(self, *other->attributes);
    other->dieWithAttacker(self, Creature::DropType::ONLY_INVENTORY);
    self->spendTime(2);
  });
}

WItem Creature::getWeapon() const {
  vector<WItem> it = equipment->getSlotItems(EquipmentSlot::WEAPON);
  if (it.empty())
    return nullptr;
  else
    return it.getOnlyElement();
}

CreatureAction Creature::applyItem(WItem item) const {
  if (!contains({ItemClass::TOOL, ItemClass::POTION, ItemClass::FOOD, ItemClass::BOOK, ItemClass::SCROLL},
      item->getClass()) || !getBody().isHumanoid())
    return CreatureAction("You can't apply this item");
  if (getBody().numGood(BodyPart::ARM) == 0)
    return CreatureAction("You have no healthy arms!");
  return CreatureAction(this, [=] (WCreature self) {
      double time = item->getApplyTime();
      playerMessage("You " + item->getApplyMsgFirstPerson(self));
      monsterMessage(getName().the() + " " + item->getApplyMsgThirdPerson(self), item->getNoSeeApplyMsg());
      item->apply(self);
      if (item->isDiscarded()) {
        self->equipment->removeItem(item, self);
      }
      self->spendTime(time);
  });
}

CreatureAction Creature::throwItem(WItem item, Vec2 direction) const {
  if (!getBody().numGood(BodyPart::ARM) || !getBody().isHumanoid())
    return CreatureAction("You can't throw anything!");
  else if (item->getWeight() > 20)
    return CreatureAction(item->getTheName() + " is too heavy!");
  int dist = 0;
  int accuracyVariance = 10;
  int attackVariance = 10;
  int str = getAttr(AttrType::STRENGTH);
  if (item->getWeight() <= 0.5)
    dist = 10 * str / 15;
  else if (item->getWeight() <= 5)
    dist = 5 * str / 15;
  else if (item->getWeight() <= 20)
    dist = 2 * str / 15;
  else 
    FATAL << "Item too heavy.";
  int accuracy = Random.get(-accuracyVariance, accuracyVariance) +
      getModifier(ModifierType::THROWN_ACCURACY) + item->getModifier(ModifierType::THROWN_ACCURACY);
  int damage = Random.get(-attackVariance, attackVariance) +
      getModifier(ModifierType::THROWN_DAMAGE) + item->getModifier(ModifierType::THROWN_DAMAGE);
  if (item->getAttackType() == AttackType::STAB) {
    damage += Skill::get(SkillId::KNIFE_THROWING)->getModifier(this, ModifierType::THROWN_DAMAGE);
    accuracy += Skill::get(SkillId::KNIFE_THROWING)->getModifier(this, ModifierType::THROWN_ACCURACY);
  }
  return CreatureAction(this, [=](WCreature self) {
    Attack attack(self, Random.choose(getBody().getAttackLevels()), item->getAttackType(), accuracy, damage, false,
        none);
    playerMessage("You throw " + item->getAName(false, this));
    monsterMessage(getName().the() + " throws " + item->getAName());
    self->getPosition().throwItem(self->equipment->removeItem(item, self), attack, dist, direction, getVision());
    self->spendTime(1);
  });
}

bool Creature::canSee(WConstCreature c) const {
  if (!c->getPosition().isSameLevel(position))
    return false;
  for (auto vision : creatureVisions)
    if (vision->canSee(this, c))
      return true;
  return !isBlind() && (!c->isAffected(LastingEffect::INVISIBLE) || isFriend(c)) &&
         (!c->isHidden() || c->knowsHiding(this)) && c->getPosition().isVisibleBy(this);
}

bool Creature::canSee(Position pos) const {
  return !isBlind() && pos.isVisibleBy(this);
}

bool Creature::canSee(Vec2 pos) const {
  return !isBlind() && position.withCoord(pos).isVisibleBy(this);
}
  
bool Creature::isPlayer() const {
  return getController()->isPlayer();
}

const CreatureName& Creature::getName() const {
  return attributes->getName();
}

CreatureName& Creature::getName() {
  return attributes->getName();
}

const char* Creature::identify() const {
  return getName().bare().c_str();
}

TribeSet Creature::getFriendlyTribes() const {
  if (WGame game = getGame())
    return game->getTribe(tribe)->getFriendlyTribes();
  else
    return TribeSet().insert(tribe);
}

MovementType Creature:: getMovementType() const {
  return MovementType(getFriendlyTribes(), {
      true,
      isAffected(LastingEffect::FLYING),
      attributes->getSkills().hasDiscrete(SkillId::SWIMMING),
      getBody().canWade()})
    .setForced(isBlind() || getHoldingCreature() || forceMovement)
    .setFireResistant(isAffected(LastingEffect::FIRE_RESISTANT))
    .setSunlightVulnerable(getBody().isSunlightVulnerable() && !isAffected(LastingEffect::DARKNESS_SOURCE)
        && (!getGame() || getGame()->getSunlightInfo().getState() == SunlightState::DAY));
}

int Creature::getDifficultyPoints() const {
  difficultyPoints = max<double>(difficultyPoints,
      getModifier(ModifierType::DEFENSE) + getModifier(ModifierType::ACCURACY) + getModifier(ModifierType::DAMAGE)
      + getAttr(AttrType::SPEED) / 10);
  CHECK(difficultyPoints >=0 && difficultyPoints < 100000) << getModifier(ModifierType::DEFENSE) << " "
     << getModifier(ModifierType::ACCURACY) << " " << getModifier(ModifierType::DAMAGE) << " "
     << getAttr(AttrType::SPEED) << " " << getName().bare();
  return difficultyPoints;
}

CreatureAction Creature::continueMoving() {
  if (shortestPath && shortestPath->isReachable(getPosition()))
    return move(shortestPath->getNextMove(getPosition()));
  else
    return CreatureAction();
}

CreatureAction Creature::stayIn(WLevel level, Rectangle area) {
  if (level != getLevel() || !getPosition().getCoord().inRectangle(area)) {
    if (level == getLevel())
      for (Position v : getPosition().neighbors8(Random))
        if (v.getCoord().inRectangle(area))
          if (auto action = move(v))
            return action;
    return moveTowards(Position(area.middle(), getLevel()));
  }
  return CreatureAction();
}

CreatureAction Creature::moveTowards(Position pos, bool stepOnTile) {
  if (!pos.isValid())
    return CreatureAction();
  if (pos.isSameLevel(position))
    return moveTowards(pos, false, stepOnTile);
  else if (auto stairs = position.getStairsTo(pos)) {
    if (stairs == position)
      return applySquare(position);
    else
      return moveTowards(*stairs, false, true);
  } else
    return CreatureAction();
}

bool Creature::canNavigateTo(Position pos) const {
  MovementType movement = getMovementType();
  for (Position v : pos.neighbors8())
    if (v.isConnectedTo(position, movement))
      return true;
  return false;
}

CreatureAction Creature::moveTowards(Position pos, bool away, bool stepOnTile) {
  CHECK(pos.isSameLevel(position));
  if (stepOnTile && !pos.canEnterEmpty(this))
    return CreatureAction();
  MEASURE(
  if (!away && !canNavigateTo(pos))
    return CreatureAction();
  , "Creature Sector checking " + getName().bare() + " from " + toString(position) + " to " + toString(pos));
  //INFO << "" << getPosition().getCoord() << (away ? "Moving away from" : " Moving toward ") << pos.getCoord();
  bool newPath = false;
  bool targetChanged = shortestPath && shortestPath->getTarget().dist8(pos) > getPosition().dist8(pos) / 10;
  if (!shortestPath || targetChanged || shortestPath->isReversed() != away) {
    newPath = true;
    if (!away)
      shortestPath.reset(new LevelShortestPath(this, pos, position));
    else
      shortestPath.reset(new LevelShortestPath(this, pos, position, -1.5));
  }
  CHECK(shortestPath);
  if (shortestPath->isReachable(position))
    if (auto action = move(shortestPath->getNextMove(position)))
      return action;
  if (newPath)
    return CreatureAction();
  INFO << "Reconstructing shortest path.";
  if (!away)
    shortestPath.reset(new LevelShortestPath(this, pos, position));
  else
    shortestPath.reset(new LevelShortestPath(this, pos, position, -1.5));
  if (shortestPath->isReachable(position)) {
    Position pos2 = shortestPath->getNextMove(position);
    if (auto action = move(pos2))
      return action;
    else {
      if (!pos2.canEnterEmpty(this))
        if (auto action = destroy(getPosition().getDir(pos2), DestroyAction::Type::BASH))
          return action;
      return CreatureAction();
    }
  } else {
    //INFO << "Cannot move toward " << pos.getCoord();
    return CreatureAction();
  }
}

CreatureAction Creature::moveAway(Position pos, bool pathfinding) {
  CHECK(pos.isSameLevel(position));
  if (pos.dist8(getPosition()) <= 5 && pathfinding)
    if (auto action = moveTowards(pos, true, false))
      return action;
  pair<Vec2, Vec2> dirs = pos.getDir(getPosition()).approxL1();
  vector<CreatureAction> moves;
  if (auto action = move(dirs.first))
    moves.push_back(action);
  if (auto action = move(dirs.second))
    moves.push_back(action);
  if (moves.size() > 0)
    return moves[Random.get(moves.size())];
  return CreatureAction();
}

bool Creature::atTarget() const {
  return shortestPath && getPosition() == shortestPath->getTarget();
}

void Creature::youHit(BodyPart part, AttackType type) const {
  switch (part) {
    case BodyPart::BACK:
        switch (type) {
          case AttackType::SHOOT: you(MsgType::ARE, "shot in the back!"); break;
          case AttackType::BITE: you(MsgType::ARE, "bitten in the neck!"); break;
          case AttackType::CUT: you(MsgType::YOUR, "throat is cut!"); break;
          case AttackType::CRUSH: you(MsgType::YOUR, "spine is crushed!"); break;
          case AttackType::PUNCH: you(MsgType::YOUR, "neck is broken!"); break;
          case AttackType::HIT: you(MsgType::ARE, "hit in the back of the head!"); break;
          case AttackType::STAB: you(MsgType::ARE, "stabbed in the "_s + 
                                     Random.choose("back"_s, "neck"_s)); break;
          case AttackType::SPELL: you(MsgType::ARE, "ripped to pieces!"); break;
          default: FATAL << "Unhandled attack type " << int(type);
        }
        break;
    case BodyPart::HEAD: 
        switch (type) {
          case AttackType::SHOOT: you(MsgType::ARE, "shot in the " +
                                      Random.choose("eye"_s, "neck"_s, "forehead"_s) + "!"); break;
          case AttackType::BITE: you(MsgType::YOUR, "head is bitten off!"); break;
          case AttackType::CUT: you(MsgType::YOUR, "head is chopped off!"); break;
          case AttackType::CRUSH: you(MsgType::YOUR, "skull is shattered!"); break;
          case AttackType::PUNCH: you(MsgType::YOUR, "neck is broken!"); break;
          case AttackType::HIT: you(MsgType::ARE, "hit in the head!"); break;
          case AttackType::STAB: you(MsgType::ARE, "stabbed in the eye!"); break;
          case AttackType::SPELL: you(MsgType::YOUR, "head is ripped to pieces!"); break;
          default: FATAL << "Unhandled attack type " << int(type);
        }
        break;
    case BodyPart::TORSO:
        switch (type) {
          case AttackType::SHOOT: you(MsgType::YOUR, "shot in the heart!"); break;
          case AttackType::BITE: you(MsgType::YOUR, "internal organs are ripped out!"); break;
          case AttackType::CUT: you(MsgType::ARE, "cut in half!"); break;
          case AttackType::STAB: you(MsgType::ARE, "stabbed in the " +
                                     Random.choose("stomach"_s, "heart"_s) + "!"); break;
          case AttackType::CRUSH: you(MsgType::YOUR, "ribs and internal organs are crushed!"); break;
          case AttackType::HIT: you(MsgType::ARE, "hit in the chest!"); break;
          case AttackType::PUNCH: you(MsgType::YOUR, "stomach receives a deadly blow!"); break;
          case AttackType::SPELL: you(MsgType::ARE, "ripped to pieces!"); break;
          default: FATAL << "Unhandled attack type " << int(type);
        }
        break;
    case BodyPart::ARM:
        switch (type) {
          case AttackType::SHOOT: you(MsgType::YOUR, "shot in the arm!"); break;
          case AttackType::BITE: you(MsgType::YOUR, "arm is bitten off!"); break;
          case AttackType::CUT: you(MsgType::YOUR, "arm is chopped off!"); break;
          case AttackType::STAB: you(MsgType::ARE, "stabbed in the arm!"); break;
          case AttackType::CRUSH: you(MsgType::YOUR, "arm is smashed!"); break;
          case AttackType::HIT: you(MsgType::ARE, "hit in the arm!"); break;
          case AttackType::PUNCH: you(MsgType::YOUR, "arm is broken!"); break;
          case AttackType::SPELL: you(MsgType::YOUR, "arm is ripped to pieces!"); break;
          default: FATAL << "Unhandled attack type " << int(type);
        }
        break;
    case BodyPart::WING:
        switch (type) {
          case AttackType::SHOOT: you(MsgType::YOUR, "shot in the wing!"); break;
          case AttackType::BITE: you(MsgType::YOUR, "wing is bitten off!"); break;
          case AttackType::CUT: you(MsgType::YOUR, "wing is chopped off!"); break;
          case AttackType::STAB: you(MsgType::ARE, "stabbed in the wing!"); break;
          case AttackType::CRUSH: you(MsgType::YOUR, "wing is smashed!"); break;
          case AttackType::HIT: you(MsgType::ARE, "hit in the wing!"); break;
          case AttackType::PUNCH: you(MsgType::YOUR, "wing is broken!"); break;
          case AttackType::SPELL: you(MsgType::YOUR, "wing is ripped to pieces!"); break;
          default: FATAL << "Unhandled attack type " << int(type);
        }
        break;
    case BodyPart::LEG:
        switch (type) {
          case AttackType::SHOOT: you(MsgType::YOUR, "shot in the leg!"); break;
          case AttackType::BITE: you(MsgType::YOUR, "leg is bitten off!"); break;
          case AttackType::CUT: you(MsgType::YOUR, "leg is cut off!"); break;
          case AttackType::STAB: you(MsgType::YOUR, "stabbed in the leg!"); break;
          case AttackType::CRUSH: you(MsgType::YOUR, "knee is crushed!"); break;
          case AttackType::HIT: you(MsgType::ARE, "hit in the leg!"); break;
          case AttackType::PUNCH: you(MsgType::YOUR, "leg is broken!"); break;
          case AttackType::SPELL: you(MsgType::YOUR, "leg is ripped to pieces!"); break;
          default: FATAL << "Unhandled attack type " << int(type);
        }
        break;
  }
}

bool Creature::isUnknownAttacker(WConstCreature c) const {
  return unknownAttackers.contains(c);
}

void Creature::updateVision() {
  if (attributes->getSkills().hasDiscrete(SkillId::NIGHT_VISION))
    vision = VisionId::NIGHT;
  else if (attributes->getSkills().hasDiscrete(SkillId::ELF_VISION) || isAffected(LastingEffect::FLYING))
    vision = VisionId::ELF;
  else
    vision = VisionId::NORMAL; 
}

VisionId Creature::getVision() const {
  return vision;
}

const CreatureDebt& Creature::getDebt() const {
  return *debt;
}

CreatureDebt& Creature::getDebt() {
  return *debt;
}

void Creature::updateVisibleCreatures() {
  int range = FieldOfView::sightRange;
  visibleEnemies.clear();
  visibleCreatures.clear();
  for (WCreature c : position.getAllCreatures(range)) 
    if (canSee(c) || isUnknownAttacker(c)) {
      visibleCreatures.push_back(c->getPosition());
      if (isEnemy(c))
        visibleEnemies.push_back(c->getPosition());
    }
}

vector<WCreature> Creature::getVisibleEnemies() const {
  vector<WCreature> ret;
  for (Position p : visibleEnemies)
    if (WCreature c = p.getCreature())
      if (!c->isDead())
        ret.push_back(c);
  return ret;
}

vector<WCreature> Creature::getVisibleCreatures() const {
  vector<WCreature> ret;
  for (Position p : visibleCreatures)
    if (WCreature c = p.getCreature())
      if (!c->isDead())
        ret.push_back(c);
  return ret;
}

vector<Position> Creature::getVisibleTiles() const {
  if (isBlind())
    return {};
  else
    return getPosition().getVisibleTiles(getVision());
}

const char* getMoraleText(double morale) {
  if (morale >= 0.7)
    return "ecstatic";
  if (morale >= 0.2)
    return "merry";
  if (morale < -0.7)
    return "depressed";
  if (morale < -0.2)
    return "unhappy";
  return nullptr;
}

vector<string> Creature::getMainAdjectives() const {
  vector<string> ret;
  if (isBlind())
    ret.push_back("blind");
  if (isAffected(LastingEffect::INVISIBLE))
    ret.push_back("invisible");
  if (getBody().numBodyParts(BodyPart::ARM) == 1)
    ret.push_back("one-armed");
  if (getBody().numBodyParts(BodyPart::ARM) == 0)
    ret.push_back("armless");
  if (getBody().numBodyParts(BodyPart::LEG) == 1)
    ret.push_back("one-legged");
  if (getBody().numBodyParts(BodyPart::LEG) == 0)
    ret.push_back("legless");
  if (isAffected(LastingEffect::HALLU))
    ret.push_back("tripped");
  if (auto text = getMoraleText(getMorale()))
    ret.push_back(text);
  return ret;
}

vector<AdjectiveInfo> Creature::getWeaponAdjective() const {
  if (const WItem weapon = getWeapon())
    return {{"Wielding " + weapon->getAName(), ""}};
  else
    return {};
}

vector<AdjectiveInfo> Creature::getGoodAdjectives() const {
  vector<AdjectiveInfo> ret;
  if (!getWeapon() && !getBody().isHumanoid()) {
    ret.push_back({"+" + toString(attributes->getBarehandedDamage()) + " unarmed attack", ""});
  }
  for (LastingEffect effect : ENUM_ALL(LastingEffect))
    if (isAffected(effect))
      if (const char* name = LastingEffects::getGoodAdjective(effect)) {
        ret.push_back({ name, Effect::getDescription(effect) });
        if (!attributes->isAffectedPermanently(effect))
          ret.back().name += "  " + attributes->getRemainingString(effect, getGlobalTime());
      }
  if (getBody().isUndead())
    ret.push_back({"Undead",
        "Undead creatures don't take regular damage and need to be killed by chopping up or using fire."});
  if (morale > 0)
    if (auto text = getMoraleText(getMorale()))
      ret.push_back({capitalFirst(text),
          "Morale affects minion's productivity and chances of fleeing from battle."});
  return ret;
}

vector<AdjectiveInfo> Creature::getBadAdjectives() const {
  vector<AdjectiveInfo> ret;
  if (!getWeapon() && getBody().isHumanoid()) {
    ret.push_back({"+" + toString(attributes->getBarehandedDamage()) + " unarmed attack", ""});
  }
  getBody().getBadAdjectives(ret);
  for (LastingEffect effect : ENUM_ALL(LastingEffect))
    if (isAffected(effect))
      if (const char* name = LastingEffects::getBadAdjective(effect)) {
        ret.push_back({ name, Effect::getDescription(effect) });
        if (!attributes->isAffectedPermanently(effect))
          ret.back().name += "  " + attributes->getRemainingString(effect, getGlobalTime());
      }
  if (isBlind())
    ret.push_back({"Blind" + (isAffected(LastingEffect::BLIND) ?
          (" " + attributes->getRemainingString(LastingEffect::BLIND, getGlobalTime())) : ""), ""});
  if (morale < 0)
    if (auto text = getMoraleText(getMorale()))
      ret.push_back({capitalFirst(text),
          "Morale affects minion's productivity and chances of fleeing from battle."});
  return ret;
}

bool Creature::isSameSector(Position pos) const {
  return pos.isConnectedTo(position, getMovementType());
}

void Creature::setInCombat() {
  lastCombatTime = getGame()->getGlobalTime();
}

bool Creature::wasInCombat(double numLastTurns) const {
  return lastCombatTime && *lastCombatTime >= getGame()->getGlobalTime() - numLastTurns;
}

